@@ -28,6 +28,7 @@ import com.vizdom.util.UnreachableCodeException;
 import java.io.*;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
+import java.lang.reflect.Modifier;
 import java.math.BigDecimal;
 import java.net.Socket;
 import java.sql.*;
@@ -40,7 +41,7 @@ import java.util.Properties;
  * connection-accepting server.
  *
  * @author Gennis Emerson
- * @version $Revision: 1.30 $
+ * @version $Revision: 1.33 $
  */
 /* The handleMessage methods may throw only SQLExceptions and 
  * RuntimeExceptions, the first if they can recover, the second
@@ -166,7 +167,7 @@ class Connection implements Runnable
                     "request: " + request);
 
                 BerIdentifier id = request.getIdentifier(); 
-                Debug.assert(id.getTagClass() == BerTypes.APPLICATION);
+                Debug.assertTrue(id.getTagClass() == BerTypes.APPLICATION);
                 int tagNumber = id.getTagNumber();
                 switch (tagNumber)
                 {
@@ -563,6 +564,14 @@ class Connection implements Runnable
                 case Types.DATE:      // ??? Perhaps we could accept
                 case Types.TIME:      // ??? dates in the default JDBC
                 case Types.TIMESTAMP: // ??? format.
+                    // TODO: Documentation in the Changes file for DBI
+                    // 1.41 indicates that if a type was
+                    // explicitly supplied, we can assume the
+                    // data is in the standard form for that
+                    // type. For example, dates should be
+                    // yyyy-MM-dd. This would let us create a
+                    // Date and call setDate here.
+
                 case Types.OTHER: 
                 default: 
                     stmt.setString(i + 1, params[i].value.toString());
@@ -677,6 +686,7 @@ class Connection implements Runnable
                         row[i] = rs.getBytes(i + 1);
                         break;
                     case Types.LONGVARBINARY: 
+                    case Types.BLOB: 
                         if (longReadLen == 0)
                             row[i] = null;
                         else
@@ -685,7 +695,8 @@ class Connection implements Runnable
                                 longReadLen, longTruncOk);
                         }
                         break;
-                    case Types.LONGVARCHAR: 
+                    case Types.LONGVARCHAR:
+                    case Types.CLOB: 
                         if (longReadLen == 0)
                             row[i] = null;
                         else
@@ -1241,15 +1252,48 @@ class Connection implements Runnable
             
             try 
             {
-                // Look up the method and call it.
+                // Look up the method and call it. If the
+                // implementation class isn't public, we may need
+                // to look for the public interface which specified
+                // the method.
                 Class c = anObject.getClass();
                 if (aParameterList.length == 0)
                     parameterClasses = null;
                 Method method = c.getMethod(aMethodName, parameterClasses);
+                int mod = c.getModifiers();
+                if (!Modifier.isPublic(mod))
+                {
+                    Debug.getLogWriter(Debug.TEDIOUS).println(mThreadId + 
+                        c.getName() + " is not public; looking for a " +
+                        "public interface with method " + aMethodName);
+                    Class[] interfaces = c.getInterfaces(); 
+                    for (int i = 0; i < interfaces.length; i++)
+                    {
+                        mod = interfaces[i].getModifiers();
+                        if (Modifier.isPublic(mod))
+                        {
+                            try
+                            {
+                                method = interfaces[i].getMethod(
+                                    aMethodName, parameterClasses);
+                                Debug.getLogWriter(Debug.TEDIOUS).println(
+                                    mThreadId + "Found public interface " +
+                                    interfaces[i].getName() + 
+                                    " with method " + aMethodName); 
+                                break;
+                            }
+                            catch (NoSuchMethodException e)
+                            {
+                            }
+                            catch (SecurityException e)
+                            {
+                            }
+                        }
+                    }
+                }
                 
                 Debug.getLogWriter(Debug.TEDIOUS).println(mThreadId + 
                     "found method: " + method.toString());
-                
                 Object returnValue = null;
                 returnValue = method.invoke(anObject, parameterObjects ); 
                 // Null should be turned into BerNull. Boolean is
